/** Put here CSS that should apply to editor screens only. **/

form {
    margin-top: 16px;
}

/* deform workarounds while we wait for deform_bootstrap to become perfect */
form.deform ul {
    list-style: none;
    margin-left: 0;
}

form.deform div.input label {
    padding-top: 0;
    margin-right: 5px;
}
